package daobeauty.interceptor;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.Criteria;
import org.hibernate.MappingException;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.metadata.ClassMetadata;

import daobeauty.dao.SessionFactoryAware;

/**
 * Created 08.09.2008
 * @author M. Serhat Cinar
 *
 */
public class MethodInterceptor implements SessionFactoryAware{
	private static final Log LOG = LogFactory.getLog(MethodInterceptor.class);
	
	public static enum ReturnType{Single, List};
	
	private boolean initialized = false;
	
	private Class domainClass;
	private Class primaryKeyClass;
	private SessionFactory sessionFactory;
	private Object target;
	private final ArrayList delegates = new ArrayList();
	
	private final HashMap<String, Object> delegatesByMethodSignature = new HashMap<String, Object>();
	private final HashMap<Class, Object> delegateInstancesByDelegateClass = new HashMap<Class, Object>();
	private final HashMap<String, Method> delegateMethodsByMethodSignature = new HashMap<String, Method>();
	
	private final HashSet<String> targetMethodSignatures = new HashSet<String>();
	
	private final HashMap<String, ReturnType> returnTypesByMethodSignatures = new HashMap<String, ReturnType>();
	
	private final HashSet<String> domainProperties = new HashSet<String>();
	
	private static final HashSet<String> ignoredMethods = new HashSet<String>();
	static{
		ignoredMethods.add("getClass");
		ignoredMethods.add("wait");
		ignoredMethods.add("notify");
		ignoredMethods.add("notifyAll");
		ignoredMethods.add("toString");
		ignoredMethods.add("equals");
		ignoredMethods.add("hashCode");
		ignoredMethods.add("clone");
		ignoredMethods.add("compareTo");
		ignoredMethods.add("get");
		ignoredMethods.add("set");
		ignoredMethods.add("add");
		ignoredMethods.add("addAll");
	}
	
	public MethodInterceptor(){}
	
	public MethodInterceptor(SessionFactory sessionFactory, Class domainClass){
		this.sessionFactory = sessionFactory;
		this.domainClass = domainClass;
	}
	
	public MethodInterceptor(SessionFactory sessionFactory, Class domainClass, Object target){
		this.sessionFactory = sessionFactory;
		this.domainClass = domainClass;
		this.target = target;
	}
		
	@SuppressWarnings("unchecked")
	public void initialize(){
		if (LOG.isDebugEnabled()){
			LOG.debug("initialize");
		}
		if (initialized){
			throw new IllegalStateException(getClass().getName()+" has already been initialized");
		}
		if (sessionFactory==null){
			throw new IllegalArgumentException("sessionFactory may not be null");
		}

		if (domainClass==null){
			throw new IllegalArgumentException("domainClass may not be null");
		}
		if (LOG.isDebugEnabled()){
			LOG.debug("domain class: "+domainClass);
		}
		
		final ClassMetadata cmd = sessionFactory.getClassMetadata(domainClass);
		primaryKeyClass = cmd.getIdentifierType().getReturnedClass();
		domainProperties.add(cmd.getIdentifierPropertyName());
		if (LOG.isDebugEnabled()){
			LOG.debug("hibernate identifier property: "+primaryKeyClass);
		}
		final String [] propertyNames = cmd.getPropertyNames();
		if (propertyNames!=null){
			for (int i=0; i<propertyNames.length; i++){
				domainProperties.add(propertyNames[i]);
				LOG.debug("hibernate property: "+propertyNames[i]);
			}
		}
		
		returnTypesByMethodSignatures.clear();

		Method methods[] = domainClass.getMethods();
		String methodName;

		
		// delegates
		
		delegatesByMethodSignature.clear();
		delegateInstancesByDelegateClass.clear();
		// analyse and cache
		if (delegates!=null){
			Class returnType;
			String methodSignature;
			for (Object delegate:delegates){
				methods = delegate.getClass().getMethods();
				if (methods!=null){
					for (int m=0; m<methods.length; m++){
						methodName = methods[m].getName();
						if (!ignoredMethods.contains(methodName)){
							methodSignature = MethodSignature.toMethodSignatureString(methods[m]);
							delegatesByMethodSignature.put(methodSignature, delegate);
							delegateInstancesByDelegateClass.put(delegate.getClass(), delegate);
							delegateMethodsByMethodSignature.put(methodSignature, methods[m]);
							returnType = methods[m].getReturnType();
							if (List.class.isAssignableFrom(returnType)){
								returnTypesByMethodSignatures.put(methodSignature, ReturnType.List);
							}
							else{
								returnTypesByMethodSignatures.put(methodSignature, ReturnType.Single);
							}
						}
					}
				}
			}
		}
		
		// target
		if (target!=null){
			targetMethodSignatures.clear();
			methods = target.getClass().getMethods();
			Class returnType;
			String methodSignature;
			if (methods!=null){
				for (int m=0; m<methods.length; m++){
					if (!ignoredMethods.contains(methods[m].getName())){
						returnType = methods[m].getReturnType();
						methodSignature = MethodSignature.toMethodSignatureString(methods[m]);
						targetMethodSignatures.add(methodSignature);
						if (List.class.isAssignableFrom(returnType)){
							returnTypesByMethodSignatures.put(methodSignature, ReturnType.List);
						}
						else{
							returnTypesByMethodSignatures.put(methodSignature, ReturnType.Single);
						}
					}
				}
			}
		}
		
		initialized = true;
	}
	
	/**
	 * @return the domainClass
	 */
	public Class getDomainClass() {
		return domainClass;
	}

	/**
	 * @param domainClass the domainClass to set
	 */
	public void setDomainClass(Class domainClass) {
		if (initialized){
			throw new IllegalStateException(getClass().getName()+" has already been initialized");
		}
		this.domainClass = domainClass;
	}

	/**
	 * @return the primaryKeyClass
	 */
	public Class getPrimaryKeyClass() {
		return primaryKeyClass;
	}
	
	public void setTarget(Object target){
		if (initialized){
			throw new IllegalStateException(getClass().getName()+" has already been initialized");
		}
		this.target = target;
	}
	
	public Object getTaget(){
		return target;
	}
	
	@SuppressWarnings("unchecked")
	public void setDelegates(Object [] delegates){
		if (initialized){
			throw new IllegalStateException(getClass().getName()+" has already been initialized");
		}
		this.delegates.clear();
		if (delegates!=null && delegates.length>0){
			for (int i=0; i<delegates.length; i++){
				this.delegates.add(delegates[i]);
			}
		}
	}
	
	@SuppressWarnings("unchecked")
	public void setDelegates(Collection delegates){
		if (initialized){
			throw new IllegalStateException(getClass().getName()+" has already been initialized");
		}
		this.delegates.clear();
		if (delegates!=null){
			this.delegates.addAll(delegates);
		}
	}
	
	public Collection getDelegates(){
		return delegates;
	}
	
	
	
	
	
	public Object invoke(Method method, Object[] args) throws Throwable{
		if (!initialized){
			throw new IllegalStateException(getClass().getName()+" was not initialized");
		}
		
		final String methodSignature = MethodSignature.toMethodSignatureString(method);
		final String methodName = method.getName();

		if (LOG.isDebugEnabled()){
			LOG.debug("invoke("+methodSignature+", "+args+")");
		}
		
		// 1. prüfen, ob methode durch target implementiert wird
		if (targetMethodSignatures.contains(methodSignature)){
			if (LOG.isDebugEnabled()){
				LOG.debug("found method on target");
			}
			return method.invoke(target, args);
		}
		
		// 2. prüfen, ob delegatemethode existiert
		final Object delegate = delegatesByMethodSignature.get(methodSignature);
		if (delegate!=null){
			if (LOG.isDebugEnabled()){
				LOG.debug("found delegate with method signature");
			}
			if (SessionFactoryAware.class.isAssignableFrom(delegate.getClass())){
				final SessionFactoryAware sfa = (SessionFactoryAware) delegate;
				if (sfa.getSessionFactory()==null){
					sfa.setSessionFactory(sessionFactory);
				}
			}
			final Method delegateMethod = delegateMethodsByMethodSignature.get(methodSignature);
			// invoke delegate method
			return delegateMethod.invoke(delegate, args);
		}
		
		// 3. sonderfälle: crud
		if ("save".equals(methodName)){
			if (args==null || args.length<1){
				throw new IllegalArgumentException("\"save\" without parameters invoked");
			}
			if (Collection.class.isAssignableFrom(args[0].getClass())){
				save((Collection) args[0]);
			}
			else{
				save(args[0]);
			}
			return null;
		}
		
		if ("update".equals(methodName)){
			if (args==null || args.length<1){
				throw new IllegalArgumentException("\"update\" without parameters invoked");
			}
			if (Collection.class.isAssignableFrom(args[0].getClass())){
				update((Collection) args[0]);
			}
			else{
				update(args[0]);
			}
			return null;
		}

		if ("delete".equals(methodName)){
			if (args==null || args.length<1){
				throw new IllegalArgumentException("\"delete\" without parameters invoked");
			}
			if (Collection.class.isAssignableFrom(args[0].getClass())){
				delete((Collection) args[0]);
			}
			else{
				delete(args[0]);
			}
			return null;
		}
		
		if ("saveOrUpdate".equals(methodName)){
			if (args==null || args.length<1){
				throw new IllegalArgumentException("\"saveOrUpdate\" without parameters invoked");
			}
			if (Collection.class.isAssignableFrom(args[0].getClass())){
				saveOrUpdate((Collection) args[0]);
			}
			else{
				saveOrUpdate(args[0]);
			}
			return null;
		}
		
		// 4. prüfen, ob eine namedquery existiert
		// a. namedquery speziell für domain
		Query namedQuery = getNamedQuery(domainClass, methodName);
		// b. namedquery generell
		if (namedQuery==null){
			namedQuery = getNamedQuery(methodName);
		}
		if (namedQuery!=null){
			if (LOG.isDebugEnabled()){
				LOG.debug("found named query with method signature");
			}

			fillQuery(namedQuery, args);
			return returnQueryResultByReturnType(namedQuery, methodSignature);
		}		
		
		// 5. methodennamen mit "findBy" und "count" parsen
		return new MethodNameToHibernateParser().create(getCurrentSession(), domainClass, domainProperties, methodName, args, getReturnTypeByMethodSignature(methodSignature));
	}

	private void saveOrUpdate(Object o){
		getCurrentSession().saveOrUpdate(o);
	}
	
	private void saveOrUpdate(Collection col){
		for (Object o:col){
			saveOrUpdate(o);
		}
	}
	
	private void delete(Object o){
		getCurrentSession().delete(o);
	}
	
	private void delete(Collection col){
		for (Object o:col){
			delete(o);
		}
	}
	
	private void update(Object o){
		getCurrentSession().update(o);
	}
	
	private void update(Collection col){
		for (Object o:col){
			update(o);
		}
	}
	
	private void save(Object o){
		getCurrentSession().save(o);
	}
	
	private void save(Collection col){
		for (Object o:col){
			save(o);
		}
	}
	
	private ReturnType getReturnTypeByMethodSignature(String methodSignature){
		ReturnType returnType = returnTypesByMethodSignatures.get(methodSignature);
		if (returnType==null){
			returnType = ReturnType.List;
		}
		return returnType;
	}
	
	private Object returnQueryResultByReturnType(Query query, String methodSignature){
		ReturnType rt = returnTypesByMethodSignatures.get(methodSignature);
		if (rt==null){
			rt = ReturnType.List;
		}
		switch(rt){
			case List:
			default:
				return query.list();
			case Single:
				return query.uniqueResult();
		}
	}
	
	private Query fillQuery(Query query, Object [] args){
		if (args!=null && args.length>0){
			// möglichkeiten
			// 1. args ist eine map -> key values
			if (args[0]!=null && Map.class.isAssignableFrom(args[0].getClass())){
				final Map argsMap = (Map) args[0];
				String key;
				Object value;
				for (Object argKey:argsMap.keySet()){
					key = String.valueOf(argKey);
					value = argsMap.get(argKey);
					query.setParameter(key, value);
				}
			}
			// 2. args sind objekte
			else{
				for (int i=0; i<args.length; i++){
					query.setParameter(i, args[i]);
				}
			}
		}
		return query;
	}
	
	private Criteria createCriteria(){
		return getCurrentSession().createCriteria(domainClass);
	}
	
	private Query getNamedQuery(Class domain, String queryName){
		if (LOG.isDebugEnabled()){
			LOG.debug("getNamedQuery("+domain.getName()+", "+queryName+")");
		}
		return getNamedQuery(domain.getName()+"."+queryName);
	}
	
	private Query getNamedQuery(String queryName){
		if (LOG.isDebugEnabled()){
			LOG.debug("getNamedQuery("+queryName+")");
		}
		try{
			return getCurrentSession().getNamedQuery(queryName);
		}
		catch(MappingException e){
			return null;
		}
	}

	/* (non-Javadoc)
	 * @see daobeauty.dao.hibernate.SessionFactoryAware#getCurrentSession()
	 */
	public Session getCurrentSession() {
		if (!initialized){
			throw new IllegalStateException(getClass().getName()+" has not been initialized");
		}
		return sessionFactory.getCurrentSession();
	}

	/* (non-Javadoc)
	 * @see daobeauty.dao.hibernate.SessionFactoryAware#getSessionFactory()
	 */
	public SessionFactory getSessionFactory() {
		return sessionFactory;
	}

	/* (non-Javadoc)
	 * @see daobeauty.dao.hibernate.SessionFactoryAware#setSessionFactory(org.hibernate.SessionFactory)
	 */
	public void setSessionFactory(SessionFactory sessionFactory) {
		if (initialized){
			throw new IllegalStateException(getClass().getName()+" has already been initialized");
		}
		this.sessionFactory = sessionFactory;
	}
	
	public static final MethodInterceptor create(SessionFactory sessionFactory, Class domainClass, Object target, Object[] delegates){
		final MethodInterceptor methodInterceptor = new MethodInterceptor();
		methodInterceptor.setDelegates(delegates);
		methodInterceptor.setDomainClass(domainClass);
		methodInterceptor.setSessionFactory(sessionFactory);
		methodInterceptor.setTarget(target);
		return methodInterceptor;
	}
}
